
"""
Timer Context Manager.
"""
from time import perf_counter


class Timer(object):
    """
    Create a Timer object.

    A context manager Timer object.

    Attributes
    ----------
    elapsed_time : int
        The elapsed time of the timer.
    """

    def __init__(self):
        """
        
        """
        self.elapsed_time = 0
        self._in_progress = None

    def __enter__(self):
        """Enter the runtime context and print status.
        """
        self._start = perf_counter()
        self._in_progress = True
        self.print_status()
        return self
    
    def __exit__(self, exc_type, exc_value, exc_traceback):
        """bool: Exit the runtime context and print status."""
        self._stop = perf_counter()
        self._in_progress = False
        self.elapsed_time = self._stop - self._start
        self.print_status()
        return False

    def print_status(self):
        """Print the current status."""
        if self._in_progress is None:
            print('Ready')
        elif self._in_progress:
            print('Running...')
        elif not self._in_progress:
            t = format(self.elapsed_time, '.2f')
            print(f'Finished in {t} seconds')

    @property
    def is_finished(self):
        """bool: Return True if timer has finished."""
        return self._in_progress is not None and not self._in_progress


"""

Parse GeoJSON from image metadata.

<br>


Quick Start
-----------


Import im2geojson and initialise by passing the directory of images:


```python
>>> from im2geojson import ImageToGeoJSON

# For example: if your current directory is named `parent` and this contains a folder of images: `my_images`,
# Initialise with `input_directory`:
>>> input_directory = './my_images'
>>> im2geojson = ImageToGeoJSON(input_directory=input_directory)

# Start image processing:
>>> im2geojson.start()
```
```s
Running...
Finished in 0.31 seconds
```
<br>


Summary
-------

```python
>>> im2geojson.summary
```
```s
'1 out of 6 images processed successfully'
```
<br>


Output
------

```json
// my_images.geojson
{
    "type": "FeatureCollection", 
    "title": "my_images", 
    "features": 
    [
        {
            "type": "Feature", 
            "geometry": 
            {
                "type": "Point", 
                "coordinates": [115.095269, -8.631053]
            }, 
            "properties": 
            {
                "datetime": "2023-05-05 06:19:24", 
                "filename": "EXIF.jpg"
            }
        }
    ], 
    "properties": 
    {
        "parent": "parent"
    }
}
```
<br>


Errors
------

```python
>>> im2geojson.error_dictionary
```
```s
{'my_images/MISSING_EXIF.jpg': 'AttributeError: image does not have attribute gps_latitude',
 'my_images/MISSING_DATETIME.jpg': 'AttributeError: image does not have attribute datetime_original',
 'my_images/CORRUPTED_DATETIME.jpg': "ValueError: time data 'corrupted' does not match format '%Y:%m:%d %H:%M:%S'",
 'my_images/CORRUPTED_EXIF.jpg': 'ValueError: Invalid GPS Reference X, Expecting N, S, E or W',
 'my_images/NO_EXIF.jpg': "'No metadata.'"}
```
<br>
<br>

   
***

<br>

"""

import logging

logging.getLogger('im2geojson').addHandler(logging.NullHandler())


from im2geojson.im2geojson import ImageToGeoJSON

__all__ = ['ImageToGeoJSON']




"""
Convert latitude / longitude DMS representation to decimal.
"""
from decimal import Decimal, getcontext
getcontext().prec = 9

SIX_PLACES = Decimal(10) **-6 
NORTH_REF = 'N'
SOUTH_REF = 'S'
EAST_REF = 'E'
WEST_REF = 'W'
MAX_MINUTES = 59
MAX_SECONDS = 60
MAX_LAT_DEGREES = 90
MAX_LONG_DEGREES = 180


def is_latitude(ref):
    """bool: Return True if `ref` is 'N' or 'S'."""
    return (ref == NORTH_REF or ref == SOUTH_REF)

def is_longitude(ref):
    """bool: Return True if `ref` is 'E' or 'W'."""
    return (ref == EAST_REF or ref == WEST_REF)

def dms_to_decimal(deg, min, sec, ref):
    """
    Convert degrees, minutes, seconds and reference to decimal.

    Parameters
    ----------
    deg : float
        Degrees.
    min : float
        Minutes.
    sec : float
        Seconds 
    ref : str
        The compass reference.
 
    Returns
    -------
    float
        Decimal degree as float.

    Raises
    ------
    ValueError
        Invalid GPS Reference, Expecting N, S, E or W.
        Invalid Seconds, Should be positive and less than 60.
        Invalid Minutes, Should be positive and less than 60.
        Invalid Degrees, Should be positive.
        Invalid Latitude, cannot be greater than 90 degrees.
        Invalid Longitude, cannot be greater than 180 degrees.
    """
    if ref not in [NORTH_REF, SOUTH_REF, EAST_REF, WEST_REF]:
        raise ValueError(f'ValueError: Invalid GPS Reference {ref}, Expecting N, S, E or W')
    
    if sec >= MAX_SECONDS or sec < 0:
        raise ValueError(f'ValueError: Invalid Seconds {str(sec)}, Should be positive and less than 60')
    
    if min > MAX_MINUTES or min < 0:
        raise ValueError(f'ValueError: Invalid Minutes {str(min)}, Should be positive and less than 60')
    
    if deg < 0:
        raise ValueError(f'ValueError: Invalid Degrees {str(sec)}, Should be positive')
    
    if is_latitude(ref) and deg > MAX_LAT_DEGREES:
        raise ValueError(f'ValueError: Invalid Latitude {str(deg) + ref}, cannot be greater than 90 degrees')
    
    elif is_longitude(ref) and deg > MAX_LONG_DEGREES:
        raise ValueError(f'ValueError: Invalid Longitude {str(deg) + ref}, cannot be greater than 180 degrees')
    
    sign = (-1 if ref == SOUTH_REF or ref == WEST_REF else 1)
    dec_deg = (Decimal(deg) + Decimal(min)/Decimal(60) + Decimal(sec)/Decimal(3600)) * sign

    return float(dec_deg.quantize(SIX_PLACES))



"""
Logging configuration for CLI.
"""

import logging


logger = logging.getLogger('im2geojson')
logger.setLevel(logging.DEBUG)

# console handler
console_handler = logging.StreamHandler()
console_handler.setLevel(logging.INFO) 

# formatter
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
console_handler.setFormatter(formatter)

# add console_handler to logger
logger.addHandler(console_handler)

# log configured
logger.info('logging configured')


"""
Tests for exif_reader
"""

import unittest
import os

from exif import Image

from im2geojson.exif_reader import read_exif


class TestExif(unittest.TestCase):

    def setUp(self):
        self.in_path = 'tests/test_files/test_images/test_exif/'
        self.file_dir = 'test_folder/EXIF.jpg'
        self.filepath = os.path.join(self.in_path, self.file_dir)

    def test_read_exif_coord_lat_long(self):
        lat = -8.631053
        long = 115.095269
        test_coord = (lat, long)
        coord, props, image_b, thumb_b = read_exif(self.filepath, get_image=False, get_thumbnail=False)
        self.assertEqual(test_coord, coord)

    def test_read_exif_datetime(self):
        datetime = "2023-05-05 06:19:24"
        coord, props, image_b, thumb_b = read_exif(self.filepath, get_image=False, get_thumbnail=False)
        self.assertEqual(props['datetime'], datetime)

    def test_read_exif_thumbnail_file(self):
        coord, props, image_b, thumb_b = read_exif(self.filepath, get_image=False, get_thumbnail=True)
        self.assertIsNotNone(thumb_b)
        self.assertIsNone(image_b)

    def test_read_exif_image_file(self):
        coord, props, image_b, thumb_b = read_exif(self.filepath, get_image=True, get_thumbnail=False)
        self.assertIsNotNone(image_b)
        self.assertIsNone(thumb_b)
    
    def test_read_exif_image_file_strips_exif_gps_data(self):
        coord, props, image_b, thumb_b = read_exif(self.filepath, get_image=True, get_thumbnail=False)
        image = Image(image_b)
        with self.assertRaises(AttributeError):
            image.gps_latitude

    def test_read_exif_logs_exception_for_unknown_file(self):
        file_dir = 'test_folder/NO_EXIST.jpg'
        filepath = os.path.join(self.in_path, file_dir)
        with self.assertLogs() as captured:
            read_exif(filepath, get_image=False, get_thumbnail=False)
        self.assertEqual(len(captured.records), 1)
        self.assertIn('No such file or directory', captured.records[0].getMessage())


class TestExifFromImageTypes(unittest.TestCase):

    def setUp(self):
        self.in_path = 'tests/test_files/test_images/test_multiple_image_types/'

    def test_jpeg(self):
        file_dir = 'test_folder/EXIF_jpeg.jpeg'
        filepath = os.path.join(self.in_path, file_dir)
        coord, props, image_b, thumb_b = read_exif(filepath, get_image=True, get_thumbnail=False)
        self.assertIsNotNone(image_b)

    @unittest.expectedFailure
    def test_tiff(self):
        file_dir = 'test_folder/EXIF_tiff.tiff'
        filepath = os.path.join(self.in_path, file_dir)
        coord, props, image_b, thumb_b = read_exif(filepath, get_image=True, get_thumbnail=False)
        self.assertIsNotNone(image_b)


class TestNoExif(unittest.TestCase):
    
    def setUp(self):
        self.in_path = 'tests/test_files/test_images/test_no_exif/'
        self.file_dir = 'test_folder/NO_EXIF.jpg'
        self.filepath = os.path.join(self.in_path, self.file_dir)

    def test_read_no_exif(self):
        with self.assertRaises(KeyError) as e:
            coord, props, image_b, thumb_b = read_exif(self.filepath, get_image=False, get_thumbnail=False)
        self.assertEqual("'KeyError: No metadata.'", str(e.exception))


class TestMissingExif(unittest.TestCase):
    
    def setUp(self):
        self.in_path = 'tests/test_files/test_images/test_missing_exif/'
        self.file_dir = 'test_folder/MISSING_EXIF.jpg'
        self.filepath = os.path.join(self.in_path, self.file_dir)

    def test_read_missing_exif(self):
        with self.assertRaises(AttributeError) as e:
            coord, props, image_b, thumb_b = read_exif(self.filepath, get_image=False, get_thumbnail=False)
        self.assertEqual('AttributeError: image does not have attribute gps_latitude', str(e.exception))


class TestMissingDatetime(unittest.TestCase):
    
    def setUp(self):
        self.in_path = 'tests/test_files/test_images/test_missing_datetime/'
        self.file_dir = 'test_folder/MISSING_DATETIME.jpg'
        self.filepath = os.path.join(self.in_path, self.file_dir)

    def test_read_missing_datetime(self):
        with self.assertRaises(AttributeError) as e:
            coord, props, image_b, thumb_b = read_exif(self.filepath, get_image=False, get_thumbnail=False)
        self.assertEqual('AttributeError: image does not have attribute datetime_original', str(e.exception))


class TestCorruptedGPSExif(unittest.TestCase):
    
    def setUp(self):
        self.in_path = 'tests/test_files/test_images/test_corrupted_exif/'
        self.file_dir = 'test_folder/CORRUPTED_EXIF.jpg'
        self.filepath = os.path.join(self.in_path, self.file_dir)

    def test_read_corrupted_gps(self):
        with self.assertRaises(ValueError) as e:
            coord, props, image_b, thumb_b = read_exif(self.filepath)
        self.assertEqual('ValueError: Invalid GPS Reference X, Expecting N, S, E or W', str(e.exception))


class TestCorruptedDatetime(unittest.TestCase):
    
    def setUp(self):
        self.in_path = 'tests/test_files/test_images/test_corrupted_datetime/'
        self.file_dir = 'test_folder/CORRUPTED_DATETIME.jpg'
        self.filepath = os.path.join(self.in_path, self.file_dir)

    def test_read_corrupted_datetime(self):
        with self.assertRaises(ValueError) as e:
            coord, props, image_b, thumb_b = read_exif(self.filepath, get_image=False, get_thumbnail=False)
        self.assertEqual("ValueError: time data 'corrupted' does not match format '%Y:%m:%d %H:%M:%S'", str(e.exception))



if __name__ == '__main__':
    unittest.main()             # pragma: no cover


"""
Tests for geojson_parser
"""

import unittest
from im2geojson.geojson_parser import GeoJSONParser
import geojson


class TestGeoJSONParser(unittest.TestCase):

    def test_init_geojson_parser_collections_dictionary(self):
        geojson_parser = GeoJSONParser()
        self.assertEqual({}, geojson_parser._collections_dict)

    def test_add_first_feature(self):
        geojson_parser = GeoJSONParser()
        test_title = 'Test_Title'
        geojson_parser.add_feature(
            collection_title = test_title, lat=0, long=0, properties={}
        )
        self.assertTrue(test_title in geojson_parser._collections_dict)
        self.assertEqual(1, len(geojson_parser._collections_dict[test_title]['features']))

    def test_feature_properties(self):
        geojson_parser = GeoJSONParser()
        test_title = 'Test_Title'
        test_filename = 'Test_Filename.jpg'
        geojson_parser.add_feature(
            collection_title = test_title, lat=0, long=0, properties={ 'filename': test_filename }
        )
        self.assertTrue(test_title in geojson_parser._collections_dict)
        feature = geojson_parser._collections_dict[test_title]['features'][0]
        properties = feature['properties']
        self.assertEqual(test_filename, properties['filename'])

    def test_point_feature_coord_long_lat(self):
        geojson_parser = GeoJSONParser()
        test_title = 'Test_Title'
        test_lat = -8
        test_long = 115
        geojson_parser.add_feature(
            collection_title = test_title, lat=test_lat, long=test_long, properties={}
        )
        feature = geojson_parser._collections_dict[test_title]['features'][0]
        geometry = feature['geometry']
        coords = geometry['coordinates']
        self.assertEqual([test_long, test_lat], coords)

    def test_add_second_feature(self):
        geojson_parser = GeoJSONParser()
        test_title = 'Test_Title'
        geojson_parser.add_feature(
            collection_title = test_title, lat=0, long=0, properties={}
        )
        geojson_parser.add_feature(
            collection_title = test_title, lat=0, long=0, properties={}
        )
        self.assertEqual(2, len(geojson_parser._collections_dict[test_title]['features']))

    def test_add_second_collection(self):
        geojson_parser = GeoJSONParser()
        test_title1 = 'Test_Title1'
        test_title2 = 'Test_Title2'
        geojson_parser.add_feature(
            collection_title = test_title1, lat=0, long=0, properties={}
        )
        geojson_parser.add_feature(
            collection_title = test_title2, lat=0, long=0, properties={}
        )
        self.assertEqual(1, len(geojson_parser._collections_dict[test_title1]['features']))
        self.assertEqual(1, len(geojson_parser._collections_dict[test_title2]['features']))

    def test_collection_properties(self):
        geojson_parser = GeoJSONParser()
        test_title = 'Test_Title'
        test_parent_folder = 'Test_Parent_Folder'
        geojson_parser.add_feature(
            collection_title = test_title, lat=0, long=0, properties={ }, collection_parent = test_parent_folder,
        )
        self.assertTrue(test_title in geojson_parser._collections_dict)
        properties = geojson_parser._collections_dict[test_title]['properties']
        self.assertEqual(test_parent_folder, properties['parent'])
        
    def test_geojson_parser_iterator(self):
        geojson_parser = GeoJSONParser()

        # test data
        test_lat = 0
        test_long = 0
        test_title1 = 'Test_Title1'
        test_title2 = 'Test_Title2'

        # add_feature
        geojson_parser.add_feature(
            collection_title = test_title1, lat=test_lat, long=test_long, properties={}
        )
        
        # test geojson
        test_point = geojson.Point((test_lat, test_long))
        test_feature = geojson.Feature(geometry=test_point, properties={})
        test_feature_collection = geojson.FeatureCollection(
            features=[test_feature],
            title=test_title1
            )
        
        # iterator
        it = iter(geojson_parser)
        title, feature_collection = next(it)
        self.assertEqual(test_title1, title)
        self.assertEqual(test_feature_collection, feature_collection)
        with self.assertRaises(StopIteration):
            next(it)


if __name__ == '__main__':
    unittest.main()             # pragma: no cover


"""
Parse Image metadata to GeoJSON.
"""
__docformat__ = "numpy"

import os
import glob
import json
import concurrent.futures
import logging

from .geojson_parser import GeoJSONParser
from .exif_reader import read_exif
from .timer import Timer

DEFAULT_OUTPUT_DIRECTORY = './assets'
GEOJSON_DIR = 'geojson'
IMAGE_DIR = 'images'

log = logging.getLogger('im2geojson')


class ImageToGeoJSON(object):
    """
    ImageToGeoJSON

    Note
    ----
    Saves the harvested metadata as geojson to 'output_directory`
    Optionally saves images without metadata and thumbnails images.
    """

    def __init__(self, 
                 input_directory, 
                 output_directory=DEFAULT_OUTPUT_DIRECTORY, 
                 save_images=False, 
                 save_thumbnails=False):
        """
        Initialise ImageToGeoJSON object.

        Initialise the object and creates `output_directory` and folders.
        
        Parameters
        ----------
        input_directory : str
            The path to the `input_directory`.
            
        output_directory : str, default './assets'
            The path to the `output_directory`.

        save_images : bool, default False
            Save images stripped of metadata to `output_directory`.

        save_thumbnails : bool, default False
            Save thumbnail images to `output_directory`.
        
        """
        
        self._input_directory = input_directory
        self._output_directory = output_directory.rstrip('/')
        self._save_images = save_images
        self._save_thumbnails = save_thumbnails

        self._geojson_parser = GeoJSONParser()
        self._timer = None
        self._error_dictionary = {}
        self._total_count = 0
        self._success_count = 0

        # Make Output Directories
        dir_paths = [self._geojson_dir_path]
        if save_images or save_thumbnails:
            dir_paths.append(self._image_dir_path)
        for path in dir_paths:
            try:
                os.makedirs(path)
            except FileExistsError:
                log.info(f"Folder {path} already exists.")
            else:
                log.info(f"Folder {path} created.")

    @property
    def input_directory(self):
        """str: Return the path to the `input_directory`."""
        return self._input_directory
    
    @property
    def output_directory(self):
        """str: Return the path to the `output_directory`."""
        return self._output_directory
        
    @property
    def summary(self):
        """str: Return the `summary` string."""
        return f'{self._success_count} out of {self._total_count} images processed successfully'
    
    @property
    def has_errors(self):
        """bool: Return `true` if `error_dictionary` contains errors."""
        return False if self._error_dictionary == {} else True
    
    @property
    def error_dictionary(self):
        """dict: Return the `error_dictionary`."""
        return self._error_dictionary

    def start(self):
        """
        Process the images from `input_directory`.

        """
        if self._timer is not None:
            raise RuntimeError('Error: Too many calls to function')
        
        with Timer() as self._timer:
            self._process_files()
            
    def _process_files(self):
        # Process image files concurrently
        files = glob.iglob(f'{self.input_directory}**/*.[Jj][Pp][Gg]')
        # TODO - **/*.@(jpg|JPG|jpeg|JPEG|gif|GIF|png|PNG) : Tests for gif, png
        with concurrent.futures.ThreadPoolExecutor() as executor:
            future_to_path = {executor.submit(self._process_image_file, filepath): filepath for filepath in files}
            for future in concurrent.futures.as_completed(future_to_path):
                filepath = future_to_path[future]
                self._total_count += 1
                try:
                    folder, coord, props = future.result()
                except Exception as e:
                    self._add_file_to_errors_with_exception_string(filepath, str(e))
                else:
                    parent = ImageToGeoJSON._parent_folder_from_filepath(filepath)
                    self._geojson_parser.add_feature(folder, *coord, props, parent)
                    self._success_count += 1

        # Save geojson
        for title, feature_collection in self._geojson_parser:
            geojson_file_path = os.path.join(self._geojson_dir_path, f'{title}.geojson')
            with open(geojson_file_path, 'w') as f:
                json.dump(feature_collection, f, indent=4)

    def _process_image_file(self, filepath):
        try:
            coord, props, image_b, thumb_b = read_exif(filepath, 
                                                       get_image=self._save_images, 
                                                       get_thumbnail=self._save_thumbnails)
        except Exception as e:
            raise e
        else:
            folder, filename = ImageToGeoJSON._folder_and_filename_from_filepath(filepath)
            props['filename'] = filename

            # image 
            if self._save_images and image_b is not None:
                rel_image_path = self._rel_image_path(filename)
                image_path = os.path.join(self.output_directory, rel_image_path)            

                with open(image_path, 'wb') as im:
                    im.write(image_b)
                    props["rel_image_path"] = rel_image_path

            # thumbnail 
            if self._save_thumbnails and thumb_b is not None:
                rel_thumbnail_path = self._rel_thumbnail_path(filename)
                thumbnail_path = os.path.join(self.output_directory, rel_thumbnail_path)

                with open(thumbnail_path, 'wb') as im:
                    im.write(thumb_b)
                    props["rel_thumbnail_path"] = rel_thumbnail_path

            return folder, coord, props
        
    def _add_file_to_errors_with_exception_string(self, filepath, exception_string):
        folder, filename = ImageToGeoJSON._folder_and_filename_from_filepath(filepath)
        key = os.path.join(folder, filename)
        self._error_dictionary[key] = exception_string

    def _output_parent_folder(self):
        """str: Return the output parent folder name."""
        head, folder = os.path.split(self._output_directory)
        return folder
    
    @property
    def _geojson_dir_path(self):
        """str: Return the path to the geojson directory."""
        return os.path.join(self.output_directory, GEOJSON_DIR)
    
    @property
    def _image_dir_path(self):
        """str: Return the path to the image directory."""
        return os.path.join(self.output_directory, IMAGE_DIR)

    def _rel_image_path(self, filename):
        """str: Return the relative path to the image filename."""
        return os.path.join(IMAGE_DIR, filename)
    
    def _rel_thumbnail_path(self, filename):
        """str: Return the relative path to the thumbnail image filename."""
        thumb_file_name = ImageToGeoJSON._thumbnail_filename(filename)
        return os.path.join(IMAGE_DIR, thumb_file_name)

    @staticmethod
    def _folder_and_filename_from_filepath(filepath):
        """tuple of str: Split the filepath and return the folder and filename."""
        head, filename = os.path.split(filepath)
        head, folder = os.path.split(head)
        return folder, filename
    
    @staticmethod
    def _parent_folder_from_filepath(filepath):
        """str: Split the filepath and return the parent folder."""
        head, filename = os.path.split(filepath)
        head, folder = os.path.split(head)
        head, parent = os.path.split(head)
        return parent
    
    @staticmethod
    def _thumbnail_filename(image_filename):
        """str: Split the image filename and return the thumbnail filename."""
        f_name, f_type  = image_filename.split('.')
        return f_name + '_thumb.' + f_type


"""
Command Line Interface for im2geojson.
"""

import argparse

from im2geojson.im2geojson import ImageToGeoJSON


def create_parser():
    """
    Creates a CLI parser.

    Returns
    -------
    parser : ArgumentParser
        The ArgumentParser with arguments added.
    """
    parser = argparse.ArgumentParser(
        argument_default=argparse.SUPPRESS,
        prog='im2geojson',
        description='Parse GeoJSON from image metadata',
        # epilog='Text at the bottom of help'
        )
    parser.add_argument(
        'input_directory', 
        help='Set the path to the images to process', 
        type=str,
        )
    parser.add_argument(
        '-o', 
        '--output_directory', 
        help='Set the output path', 
        type=str
        )
    parser.add_argument(
        '-s', 
        '--save_images', 
        help='Save Images stripped of metadata', 
        action='store_true'
        )
    parser.add_argument(
        '-t', 
        '--save_thumbnails', 
        help='Save thumbnail images', 
        action='store_true'
        )
    return parser

def parse_args_to_dict(args):
    """
    Parse `args` to a dictionary.

    Returns
    -------
    parsed_args_dict : dictionary
        Dictionary of parsed arguments.
    """
    parser = create_parser()
    parsed_args = parser.parse_args(args)
    parsed_args_dict = vars(parsed_args)
    return parsed_args_dict

def main(args=None):
    """
    Process images

    Process images from CLI, print summery and results.
    """
    parsed_args_dict = parse_args_to_dict(args)
    im2geo = ImageToGeoJSON(**parsed_args_dict)
    im2geo.start()
    print(im2geo.summary)
    if im2geo.has_errors:
        import pprint
        pprint.pp(im2geo.error_dictionary)


if __name__ == '__main__':
    main(args=None)                 # pragma: no cover


"""
Tests for timer
"""

import unittest
from time import sleep
import io
from contextlib import redirect_stdout

from im2geojson.timer import Timer


class TestTimer(unittest.TestCase):

    def test_timer_init(self):
        timer = Timer()
        self.assertEqual(0, timer.elapsed_time)

    def test_timer_status_ready(self):
        timer = Timer()
        f = io.StringIO()
        with redirect_stdout(f):
            timer.print_status()
        out = f.getvalue()

        self.assertEqual('Ready\n', out)

    def test_timer_status_running(self):
        timer = Timer()
        timer._in_progress = True

        f = io.StringIO()
        with redirect_stdout(f):
            timer.print_status()
        out = f.getvalue()

        self.assertEqual('Running...\n', out)

    def test_timer_status_finished(self):
        timer = Timer()
        timer._in_progress = False

        f = io.StringIO()
        with redirect_stdout(f):
            timer.print_status()
        out = f.getvalue()

        self.assertEqual('Finished in 0.00 seconds\n', out)

    def test_timer_context_manager(self):
        f = io.StringIO()
        with redirect_stdout(f):
            with Timer() as timer:
                sleep(0.01)
        out = f.getvalue()

        self.assertIn('Running...\nFinished in ', out)

    def test_timer_is_finished(self):
        with Timer() as timer:
            pass
        self.assertTrue(timer.is_finished)

    def test_timer_is_not_finished(self):
        timer = Timer()
        self.assertFalse(timer.is_finished)



if __name__ == '__main__':  
    unittest.main()             # pragma: no cover


"""
Read Exchangeable Image File Format (EXIF) metadata from an image.
"""
from exif import Image
from datetime import datetime
import warnings 
import threading
import logging

from .dms_conversion import dms_to_decimal

log = logging.getLogger('im2geojson')


def read_exif(filepath, get_image=False, get_thumbnail=False):
    """
    Read exif metadata from image file at `filepath`.
    
    Parameters
    ----------
    filepath : str
        The path to the image file.

    Returns
    -------
    (lat, long) : tuple of float
        The decimal latitude, longitude coordinate as a float.
    props : dictionary
        Dictionary containing the date the image was captured.
    image_b : bytes
        The image stripped of exif metadata.
    thumb_b : bytes
        The thumbnail.
    
    Raises
    ------
    KeyError
        If `image_file` has no metadata.
    AttributeError
        If `image_file` has missing metadata.
    ValueError
        If `image_file` has invalid metadata.
    FileNotFoundError
        If no file found at `filepath.
    """
    try:
        with open(filepath, 'rb') as image_file:
            image = Image(image_file)
            if not image.has_exif:
                raise KeyError('KeyError: No metadata.')

            # coord
            try:
                dms_lat = (*image.gps_latitude, image.gps_latitude_ref)
                dms_long = (*image.gps_longitude, image.gps_longitude_ref)
            except AttributeError as e:
                raise AttributeError(f'AttributeError: {e}') from e
            else:
                try:
                    lat = dms_to_decimal(*dms_lat)
                    long = dms_to_decimal(*dms_long)
                except ValueError as e:
                    raise e
            
            # datetime
            try:
                datetime_str = image.datetime_original
            except AttributeError as e:
                raise AttributeError(f'AttributeError: {e}') from e
            else:
                try:
                    datetime_object = datetime.strptime(datetime_str, '%Y:%m:%d %H:%M:%S')
                except ValueError as e:
                    raise ValueError(f'ValueError: {e}') from e

            # props 
            props = { 
                "datetime": str(datetime_object),
                }

            # delete exif data
            if get_image:
                with threading.RLock():
                    # Catch warning that not all data has been deleted:
                    with warnings.catch_warnings():
                        warnings.simplefilter('ignore')
                        image.delete_all()
                
            # files
            image_b = image.get_file() if get_image else None

            # TODO - try
            thumb_b = image.get_thumbnail() if get_thumbnail else None

            return (lat, long), props, image_b, thumb_b
        

    except FileNotFoundError as e:
        log.exception(f'FileNotFoundError: No such file or directory: {filepath}')


"""
Tests for geojsonphoto
"""

import unittest
import os
import shutil
import io
from contextlib import redirect_stdout

from im2geojson.im2geojson import *


class TestBaseClass(unittest.TestCase):

    def setUp(self):
        self.default_in_path = "./"
        self.input_directory = 'tests/test_files/test_images/test_exif/'
        self.output_directory = 'tests/assets/'
        self.geojson_dir_path = os.path.join(self.output_directory, GEOJSON_DIR)
        self.image_dir_path = os.path.join(self.output_directory, IMAGE_DIR)

    def tearDown(self):
        if os.path.isdir(self.output_directory):
            shutil.rmtree(self.output_directory)

        if os.path.isdir(DEFAULT_OUTPUT_DIRECTORY):
            shutil.rmtree(DEFAULT_OUTPUT_DIRECTORY)

class TestImageToGeoJSONInit(TestBaseClass):

    def setUp(self):
        super().setUp()

    def tearDown(self):
        super().tearDown()
    
    def test_in_path(self):
        im2geojson = ImageToGeoJSON(input_directory = self.input_directory)
        self.assertEqual(self.input_directory, im2geojson.input_directory)

    def test_no_in_path_raises_exception(self):
        with self.assertRaises(TypeError):
            im2geojson = ImageToGeoJSON()
    
    def test_out_path(self):
        im2geojson = ImageToGeoJSON(input_directory = self.input_directory, output_directory = self.output_directory)
        self.assertEqual(self.output_directory.rstrip('/'), im2geojson.output_directory)

    def test_output_folder(self):
        im2geojson = ImageToGeoJSON(input_directory = self.input_directory, output_directory = self.output_directory)
        self.assertEqual('assets', im2geojson._output_parent_folder())

    def test_init_creates_geojsonphoto_dir_path(self):
        im2geojson = ImageToGeoJSON(input_directory = self.input_directory, output_directory = self.output_directory)
        self.assertEqual(self.geojson_dir_path, im2geojson._geojson_dir_path)

    def test_init_creates_image_dir_path(self):
        im2geojson = ImageToGeoJSON(input_directory = self.input_directory, output_directory = self.output_directory)
        self.assertEqual(self.image_dir_path, im2geojson._image_dir_path)
    
    def test_init_geojsonphoto_parser(self):
        im2geojson = ImageToGeoJSON(input_directory = self.input_directory, output_directory = self.output_directory)
        self.assertTrue(im2geojson._geojson_parser)
    
    def test_default_init_im2geojson_creates_out_directories(self):
        self.assertFalse(os.path.isdir(self.output_directory))

        im2geojson = ImageToGeoJSON(input_directory = self.input_directory, 
                            output_directory = self.output_directory)

        self.assertTrue(os.path.isdir(self.geojson_dir_path))
        self.assertFalse(os.path.isdir(self.image_dir_path))
    
    def test_thumbnail_init_im2geojson_creates_out_directories(self):
        self.assertFalse(os.path.isdir(self.output_directory))

        im2geojson = ImageToGeoJSON(input_directory = self.input_directory, 
                            output_directory = self.output_directory, 
                            save_images=False, 
                            save_thumbnails=True)

        self.assertTrue(os.path.isdir(self.geojson_dir_path))
        self.assertTrue(os.path.isdir(self.image_dir_path))
    
    def test_image_init_im2geojson_creates_out_directories(self):
        self.assertFalse(os.path.isdir(self.output_directory))

        im2geojson = ImageToGeoJSON(input_directory = self.input_directory, 
                            output_directory = self.output_directory, 
                            save_images=True, 
                            save_thumbnails=False)

        self.assertTrue(os.path.isdir(self.geojson_dir_path))
        self.assertTrue(os.path.isdir(self.image_dir_path))

    def test_init_logs_geojson_directory_created(self):
        self.assertFalse(os.path.isdir(self.output_directory))

        with self.assertLogs() as captured:
            im2geojson = ImageToGeoJSON(input_directory = self.input_directory, 
                                output_directory = self.output_directory)

        self.assertEqual(len(captured.records), 1)
        self.assertEqual(captured.records[0].getMessage(), f"Folder {self.geojson_dir_path} created.")

    def test_init_logs_geojson_directory_exists(self):
        self.assertFalse(os.path.isdir(self.output_directory))
        os.makedirs(self.geojson_dir_path)

        with self.assertLogs() as captured:
            im2geojson = ImageToGeoJSON(input_directory = self.input_directory, 
                                output_directory = self.output_directory)

        self.assertEqual(len(captured.records), 1)
        self.assertEqual(captured.records[0].getMessage(), f"Folder {self.geojson_dir_path} already exists.")

    def test_init_logs_image_directory_created(self):
        self.assertFalse(os.path.isdir(self.output_directory))

        with self.assertLogs() as captured:
            im2geojson = ImageToGeoJSON(input_directory = self.input_directory, 
                            output_directory = self.output_directory, 
                            save_images=True, 
                            save_thumbnails=False)

        self.assertEqual(len(captured.records), 2)
        self.assertEqual(captured.records[1].getMessage(), f"Folder {self.image_dir_path} created.")
        

class TestImageToGeoJSONStart(TestBaseClass):

    def setUp(self):
        super().setUp()
        self.test_geojson_file_name = 'test_folder.geojson'

    def tearDown(self):
        super().tearDown()

    def test_im2geojson_start_creates_geojsonphoto_file(self):
        im2geojson = ImageToGeoJSON(input_directory = self.input_directory, 
                            output_directory = self.output_directory, 
                            save_images=False)
        im2geojson.start()

        self.assertTrue(os.path.isdir(self.geojson_dir_path))

        geojson_path = os.path.join(self.geojson_dir_path, self.test_geojson_file_name)
        with open(geojson_path, 'r') as f:
            jsn = json.load(f)
            self.assertIsNotNone(jsn['properties'])
            self.assertIsNotNone(jsn['features'][0]['properties']['datetime'])

            with self.assertRaises(KeyError):
                jsn['features'][0]['properties']['image_path']
            with self.assertRaises(KeyError):
                jsn['features'][0]['properties']['thumbnail_path']

    def test_im2geojson_start_creates_image_file(self):
        test_image_file = 'EXIF.jpg'
        im2geojson = ImageToGeoJSON(input_directory = self.input_directory, 
                            output_directory = self.output_directory, 
                            save_images=True)
        im2geojson.start()

        self.assertTrue(os.path.isdir(self.image_dir_path))

        image_path = os.path.join(self.image_dir_path, test_image_file)
        with open(image_path, 'r') as f:
            pass
        
        geojson_path = os.path.join(self.geojson_dir_path, self.test_geojson_file_name)
        with open(geojson_path, 'r') as f:
            jsn = json.load(f)
            self.assertIsNotNone(jsn['features'][0]['properties']['datetime'])
            self.assertIsNotNone(jsn['features'][0]['properties']['rel_image_path'])

            with self.assertRaises(KeyError):
                jsn['features'][0]['properties']['rel_thumbnail_path']

    def test_im2geojson_start_creates_thumbnail_file(self):
        test_thumbnail_file = 'EXIF_thumb.jpg'
        im2geojson = ImageToGeoJSON(input_directory = self.input_directory, 
                            output_directory = self.output_directory, 
                            save_images=False, 
                            save_thumbnails=True)
        im2geojson.start()

        self.assertTrue(os.path.isdir(self.image_dir_path))

        thumb_path = os.path.join(self.image_dir_path, test_thumbnail_file)
        with open(thumb_path, 'r') as f:
            pass
        
        geojson_path = os.path.join(self.geojson_dir_path, self.test_geojson_file_name)
        with open(geojson_path, 'r') as f:
            jsn = json.load(f)
            self.assertIsNotNone(jsn['features'][0]['properties']['datetime'])
            self.assertIsNotNone(jsn['features'][0]['properties']['rel_thumbnail_path'])

            with self.assertRaises(KeyError):
                jsn['features'][0]['properties']['rel_image_path']

    def test_im2geojson_start_feature_props_filename(self):
        test_thumbnail_file = 'EXIF_thumb.jpg'
        im2geojson = ImageToGeoJSON(input_directory = self.input_directory, 
                            output_directory = self.output_directory, 
                            save_images=False, 
                            save_thumbnails=False)
        im2geojson.start()

        geojson_path = os.path.join(self.geojson_dir_path, self.test_geojson_file_name)
        with open(geojson_path, 'r') as f:
            jsn = json.load(f)
            self.assertIsNotNone(jsn['features'][0]['properties']['filename'])


class TestImageToGeoJSONStatus(TestBaseClass):

    def setUp(self):
        super().setUp()

    def tearDown(self):
        super().tearDown()

    def test_in_progress_status(self):
        im2geojson = ImageToGeoJSON(input_directory = self.input_directory, 
                                 output_directory = self.output_directory, 
                                 save_images=False, 
                                 save_thumbnails=True)
        f = io.StringIO()
        with redirect_stdout(f):
            im2geojson.start()
        out = f.getvalue()
        self.assertIn('Running...\nFinished in ', out)

    def test_repeat_calls_to_start_raises_exception(self):
        im2geojson = ImageToGeoJSON(input_directory = self.input_directory, 
                                 output_directory = self.output_directory, 
                                 save_images=False, 
                                 save_thumbnails=True)
        im2geojson.start()
        with self.assertRaises(RuntimeError):
            im2geojson.start()


class TestImageToGeoJSONErrors(TestBaseClass):

    def setUp(self):
        super().setUp()

    def tearDown(self):
        super().tearDown()
    
    def test_no_errors(self):
        im2geojson = ImageToGeoJSON(input_directory = self.input_directory, 
                            output_directory = self.output_directory, 
                            save_images=False, 
                            save_thumbnails=True)
        im2geojson.start()
        self.assertEqual({}, im2geojson.error_dictionary)
    
    def test_errors(self):
        in_path = 'tests/test_files/test_images/test_no_exif/'
        im2geojson = ImageToGeoJSON(input_directory = in_path, 
                            output_directory = self.output_directory,
                            save_images=False, 
                            save_thumbnails=True)
        im2geojson.start()
        test_folder_file = 'test_folder/NO_EXIF.jpg'
        test_error_dictionary = {
            test_folder_file:"'KeyError: No metadata.'"
        }
        self.assertEqual(test_error_dictionary, im2geojson.error_dictionary)

    def test_has_errors(self):
        in_path = 'tests/test_files/test_images/test_no_exif/'
        im2geojson = ImageToGeoJSON(input_directory = in_path, 
                            output_directory = self.output_directory,
                            save_images=False, 
                            save_thumbnails=True)
        im2geojson.start()
        self.assertTrue(im2geojson.has_errors)

    def test_has_no_errors(self):
        in_path = 'tests/test_files/test_images/test_exif/'
        im2geojson = ImageToGeoJSON(input_directory = in_path, 
                            output_directory = self.output_directory,
                            save_images=False, 
                            save_thumbnails=True)
        im2geojson.start()
        self.assertFalse(im2geojson.has_errors)


class TestImageToGeoJSONSummary(TestBaseClass):

    def setUp(self):
        super().setUp()

    def tearDown(self):
        super().tearDown()

    def test_summary(self):
        im2geojson = ImageToGeoJSON(input_directory = self.input_directory, 
                            output_directory = self.output_directory, 
                            save_images=False, 
                            save_thumbnails=True)
        im2geojson.start()
        self.assertEqual('1 out of 1 images processed successfully', im2geojson.summary)
        
        
class TestFolderFilesFromPath(unittest.TestCase):

    def setUp(self):
        self.test_filename = 'image_file.jpg'
        self.test_thumb_filename = 'image_file_thumb.jpg'
        self.test_folder_name = 'folder'
        self.test_parent_folder_name = 'test_files'
        self.test_in_path = os.path.join(
            'tests/', 
            self.test_parent_folder_name, 
            self.test_folder_name, 
            self.test_filename
        )

    def test_filename_from_filepath(self):
        filename = ImageToGeoJSON._folder_and_filename_from_filepath(self.test_in_path)[1]
        self.assertEqual(self.test_filename, filename )

    def test_folder_from_filepath(self):
        folder = ImageToGeoJSON._folder_and_filename_from_filepath(self.test_in_path)[0]
        self.assertEqual(self.test_folder_name, folder)

    def test_parent_folder_from_filepath(self):
        parent_folder = ImageToGeoJSON._parent_folder_from_filepath(self.test_in_path)
        self.assertEqual(self.test_parent_folder_name, parent_folder)

    def test_thumbnail_filename_from_image_filename(self):
        thumbnail_filename = ImageToGeoJSON._thumbnail_filename(self.test_filename)
        self.assertEqual(self.test_thumb_filename, thumbnail_filename)



if __name__ == '__main__':
    unittest.main()             # pragma: no cover